@@ -1,10 +1,10 @@
 ;;; setup.el --- Helpful Configuration Macro    -*- lexical-binding: t -*-
 
-;; Copyright (C) 2021, 2022  Free Software Foundation, Inc.
+;; Copyright (C) 2021, 2022, 2023, 2024  Free Software Foundation, Inc.
 
 ;; Author: Philip Kaludercic <philipk@posteo.net>
 ;; Maintainer: Philip Kaludercic <~pkal/public-inbox@lists.sr.ht>
-;; Version: 1.3.0
+;; Version: 1.4.0
 ;; Package-Requires: ((emacs "26.1"))
 ;; Keywords: lisp, local
 ;; URL: https://git.sr.ht/~pkal/setup
@@ -24,7 +24,7 @@
 
 ;;; Commentary:
 
-;; The `setup` macro simplifies repetitive configuration patterns, by
+;; The `setup' macro simplifies repetitive configuration patterns, by
 ;; providing context-sensitive local macros in `setup' bodies.  These
 ;; macros can be mixed with regular elisp code without any issues,
 ;; allowing for flexible and terse configurations.  The list of local
@@ -37,41 +37,31 @@
 
 ;;; News:
 
-;;;; Version 1.3.0
+;;;; Version 1.4.0 (13Apr24)
 
-;; - Add new `:and' macro.
-;; - Add new `:with-function' macro and have `:hook-into' use
-;;   the context function instead of the context mode (Earl Hyatt)
-;; - Improve `:bind-into' handling of maps.
-
-;;;; Version 1.2.0
-
-;; - Remove `setup-wrap-to-demote-errors' from `setup-modifier-list'
-;; - Pull `setup-expand-local-macros'  back into `setup'
-;; - Let `:with-feature' and `:with-mode' check symbol properties to
-;;   improve context-setting guesses.
-;; - Move macros :hide-mode, :advise, :needs, :if-host and :load-from
-;;   to EmacsWiki.
-;; - Revert the indentation spec change for `setup-define'
-;; - Add :bind-into macro
-;; - Add :ensure key to `setup-define' to replace
-;;   the deprecated setup-ensure-... functions
-;; - Add `setup-bind' macro to simplify context modification.
-
-;;;; Version 1.1.0:
-
-;; - Fix quoting error in :file-match definition
-;; - Remove unnecessary check for lexical binding
-;; - Change `setup-define' indentation
-;; - Handle multiple expressions in :when-loaded
-;; - Improvements to setup docstring
-;; - Wrap `setup' forms with `with-demoted-errors'
-;; - Allow `setup' to be extended using `setup-modifier-list'
+;; - New :bind-to macro, that takes a key chord and binds the
+;;   context-relevant function.  E.g.
+;;
+;;   (setup (:package do-at-point)
+;;     (:bind-to "C-'"))
+;;
+;; - New :autoload-this creates an auto-load cookie for the
+;;   context-relevant function by connecting it to the
+;;   context-relevant feature.
+;;
+;; - Avoid calling `package-refresh-contents' in :package if it is not
+;;   necessary.  The check here was broken, as Ralf Schmitt noticed.
+;;
+;; - Earl Hyatt improved the :repeatable keyword to support a prefix
+;;   of non-repeating arguments.
+;;
+;; - Passing a map to :bind-into has been deprecated, since there is
+;;   no reliable way to deduce the feature that holds the map.  I
+;;   apologise for the inconvenience.  The functionality will remain
+;;   for a while, but I would advise rewriting affected configuration.
 
 ;;; Code:
 
-(require 'elisp-mode)
-
 (defvar setup-opts `((quit . ,(make-symbol "setup-quit")))
   "Alist defining the context for local macros.
 Context-modifying macros (`:with-feature', `:with-mode', ...)
@@ -129,7 +119,9 @@ Do not modify this variable by hand.  Instead use
               "that is immediately evaluated.")
       (fill-paragraph)
       (dolist (sym (sort (mapcar #'car setup-macros) #'string-lessp))
-        (newline 2)
+        (if (fboundp 'make-separator-line)
+            (insert "\n" (make-separator-line) "\n")
+          (newline 2))
         (let ((sig (mapcar
                     (lambda (arg)
                       (if (string-match "\\`&" (symbol-name arg))
@@ -169,6 +161,67 @@ NAME may also be a macro, if it can provide a symbol."
 ;;;###autoload
 (put 'setup 'function-documentation '(setup--make-docstring))
 
+(defun setup--ensure (ensure-spec args &optional check-len)
+  "Ensure that ARGS matches the form of ENSURE-SPEC.
+
+The symbol `kbd' means to apply the function `kbd' to the
+argument.  The symbol `func' means to sharp-quote the argument.
+The symbol `&rest' means that the remaining elements of
+ENSURE-SPEC are applied repeatedly to the remaining elements of
+ARGS.
+
+When CHECK-LEN is non-nil, check that the lengths
+of ENSURE-SPEC and ARGS are compatible."
+  (when check-len
+    (let ((check ensure-spec)
+          (found nil)
+          (count 0))
+      (while check
+        (if (eq (car check) '&rest)
+            (if (zerop (mod (- (length args) count)
+                            (length (cdr check))))
+                (setq found t
+                      check nil)
+              (error "Bad `:ensure' spec or argument list"))
+          (setq check (cdr check)
+                count (1+ count))))
+      (unless (or found (= (length args) count))
+        (error "Bad `:ensure' spec or argument list"))))
+
+  (let ((result))
+    (while args
+      (let ((ensure (pop ensure-spec)))
+        (if (eq ensure '&rest)
+            (let* ((rest-spec ensure-spec)
+                   (rest-spec-len (length rest-spec)))
+              ;; Consume remaining `args'
+              (while args
+                (dolist (ensured
+                         (setup--ensure rest-spec
+                                        (let ((to-be-ensured))
+                                          (dotimes (_ rest-spec-len)
+                                            (push (pop args) to-be-ensured))
+                                          (nreverse to-be-ensured))
+                                        nil))
+                  (push ensured result))))
+          (let ((arg (pop args)))
+            (push (cond ((null ensure) arg) ;Do not modify argument
+                        ((eq ensure 'kbd) (cond
+                                           ((stringp arg) (kbd arg))
+                                           ((symbolp arg) `(kbd ,arg))
+                                           (arg)))
+                        ((eq ensure 'func) (cond
+                                            ((eq (car-safe arg) 'function)
+                                             arg)
+                                            ((eq (car-safe arg) 'quote)
+                                             `#',(cadr arg))
+                                            ((symbolp arg)
+                                             `#',arg)
+                                            (arg)))
+                        ((error "Invalid ensure spec %S" ensure)))
+                  result)))))
+    (nreverse result)))
+
 (defun setup-define (name fn &rest opts)
   "Define `setup'-local macro NAME using function FN.
 The plist OPTS may contain the key-value pairs:
@@ -182,6 +235,10 @@ Wrap the macro in a `with-eval-after-load' body.
   :repeatable ARITY
 Allow macro to be automatically repeated.  If ARITY is t, use
 `func-arity' to determine the minimal number of arguments.
+If ARITY is a dotted pair, then the car of ARITY is the number
+of shared arguments in the repeated function calls and the
+cdr is the number of arguments that make of the remainder
+of the function call.  The cdr may also be t, as above.
 
   :signature SIG
 Give an advertised calling convention.
@@ -204,7 +261,9 @@ If not given, it is assumed nothing is evaluated.
 A list of symbols indicating what kind of argument each parameter
 to FN is.  If the nth parameter is not to be reinterpreted, the
 nth symbol in SPEC should nil.  For key bindings `kbd' and for
-functions `func'.  Any other value is invalid."
+functions `func'.  `&rest' means that the remaining symbols apply
+to the remaining arguments repeatedly.  Any other value is
+invalid."
   (declare (indent 1))
   ;; NB.: NAME is not required to by a keyword, even though all macros
   ;;      specified on the next page use keywords.  The rationale for
@@ -225,41 +284,46 @@ functions `func'.  Any other value is invalid."
   (put name 'lisp-indent-function (plist-get opts :indent))
   ;; define macro for `macroexpand-all'
   (setf (alist-get name setup-macros)   ;New in Emacs-25.
-        (let* ((arity (if (eq (plist-get opts :repeatable) t)
-                          (car (func-arity fn))
-                        (plist-get opts :repeatable)))
-               (fn (if (null arity) fn
+        (let* ((possible-num-repeated (if (eq (plist-get opts :repeatable) t)
+                                          (car (func-arity fn))
+                                        (plist-get opts :repeatable)))
+               (fn (if (null possible-num-repeated)
+                       fn
                      (lambda (&rest args)
-                       (unless (zerop (mod (length args) arity))
-                         (error "Illegal arguments"))
-                       (let (aggr)
+                       (let ((aggr)
+                             (using-shared-args (consp possible-num-repeated))
+                             (num-shared)
+                             (shared-args)
+                             (num-repeated))
+
+                         (if using-shared-args
+                             (progn
+                               (setq num-shared (car possible-num-repeated)
+                                     num-repeated (if (eq (cdr possible-num-repeated) t)
+                                                      (- (car (func-arity fn))
+                                                         num-shared)
+                                                    (cdr possible-num-repeated))
+                                     shared-args args
+                                     args (nthcdr num-shared args))
+                               (setf (nthcdr num-shared shared-args) nil))
+                           (setq num-repeated possible-num-repeated))
+
+                         (unless (zerop (mod (length args) num-repeated))
+                           (error "Illegal arguments"))
+
                          (while args
-                           (let ((rest (nthcdr arity args)))
-                             (setf (nthcdr arity args) nil)
+                           (let ((rest (nthcdr num-repeated args)))
+                             (setf (nthcdr num-repeated args) nil)
                              (let ((ensure-spec (plist-get opts :ensure)))
-                               (when ensure-spec
-                                 (dotimes (i (length args))
-                                   (let ((ensure (nth i ensure-spec))
-                                         (arg (nth i args)))
-                                     (cond
-                                      ((null ensure)) ;Do not modify argument
-                                      ((eq ensure 'kbd)
-                                       (setf (nth i args)
-                                             (cond
-                                              ((stringp arg) (kbd arg))
-                                              ((symbolp arg) `(kbd ,arg))
-                                              (arg))))
-                                      ((eq ensure 'func)
-                                       (setf (nth i args)
-                                             (cond
-                                              ((eq (car-safe arg) 'function)
-                                               arg)
-                                              ((eq (car-safe arg) 'quote)
-                                               `#',(cadr arg))
-                                              ((symbolp arg)
-                                               `#',arg)
-                                              (arg))))
-                                      ((error "Invalid ensure spec %S" ensure)))))))
+                               (cond
+                                ((and using-shared-args ensure-spec)
+                                 (setq args (setup--ensure
+                                             ensure-spec
+                                             (append shared-args args) t)))
+                                (using-shared-args
+                                 (setq args (append shared-args args)))
+                                (ensure-spec
+                                 (setq args (setup--ensure ensure-spec args t)))))
                              (push (apply fn args) aggr)
                              (setq args rest)))
                          (macroexp-progn (nreverse aggr)))))))
@@ -281,6 +345,7 @@ functions `func'.  Any other value is invalid."
 
 (defun setup--xref-def-function (symbol)
   "Return an elisp xref location for SYMBOL."
+  (require 'elisp-mode)
   (and (assq symbol setup-macros)
        (let ((file (get symbol 'setup-definition-file)))
          (list (elisp--xref-make-xref nil symbol file)))))
@@ -311,15 +376,15 @@ Each entry in VARS is a list of the form (VAR VAL), comparable to
 `let'.  This macro makes sure that the BODY is expanded correctly
 so that it can make use of the new bindings in VARS."
   (declare (debug let) (indent 1))
-  ;; The macro modifies VARS in place, inserting unquotes in the right
-  ;; places to convert a `let'-formed list into a alist.  The unquoted
-  ;; values are then handled by the backquote inserted by the macro.
-  ;; The list this generates is destructively concatenated to the
-  ;; beginning of setup-ops, which is safe because backquoting expands
-  ;; to a new list allocation.
-  (dolist (var vars)
-    (setcdr var (list '\, (cadr var))))
-  `(let ((setup-opts (nconc ,(list '\` vars) setup-opts)))
+  `(let ((setup-opts (append
+                      (list
+                       ,@(mapcar
+                          (lambda (bind)
+                            (list 'cons
+                                  (list 'quote (car bind))
+                                  (cadr bind)))
+                          vars))
+                      setup-opts)))
      (setup-expand ,body)))
 
 (defun setup-quit (&optional return)
@@ -344,6 +409,18 @@ VAL into one s-expression."
                          (if (member ,sym list)
                              list
                            (append list (list ,sym)))))))
+          ((eq (car-safe name) 'append*)
+           (funcall wrap-fn
+                    (cadr name)
+                    (let ((sym (gensym))
+                          (i (gensym)))
+                      `(let ((list ,(funcall old-val-fn (cadr name)))
+                             (,sym nil))
+                         (dolist (,i ,val)
+                           (if (member ,i list)
+                               nil
+                             (push ,i ,sym)))
+                         (append list (nreverse ,sym))))))
           ((eq (car-safe name) 'prepend)
            (funcall wrap-fn
                     (cadr name)
@@ -353,10 +430,30 @@ VAL into one s-expression."
                          (if (member ,sym list)
                              list
                            (cons ,sym list))))))
+          ((eq (car-safe name) 'prepend*)
+           (funcall wrap-fn
+                    (cadr name)
+                    (let ((sym (gensym))
+                          (i (gensym)))
+                      `(let ((list ,(funcall old-val-fn (cadr name)))
+                             (,sym nil))
+                         (dolist (,i ,val)
+                           (if (member ,i list)
+                               nil
+                             (push ,i ,sym)))
+                         (append (nreverse ,sym) list)))))
           ((eq (car-safe name) 'remove)
            (funcall wrap-fn
                     (cadr name)
                     `(remove ,val ,(funcall old-val-fn (cadr name)))))
+          ((eq (car-safe name) 'remove*)
+           (funcall wrap-fn
+                    (cadr name)
+                    (let ((i (gensym)))
+                      `(let ((list ,(funcall old-val-fn (cadr name))))
+                         (dolist (,i ,val)
+                           (setq list (remove ,i list)))
+                         list))))
           ((error "Invalid option %S" name)))))
 
 
@@ -372,12 +469,12 @@ VAL into one s-expression."
                                 (intern (format "%s-mode" feature)))))
                     (setup-bind body
                       (feature feature)
-                      (mode (or (get features 'setup-mode) mode))
-                      (func (or (get features 'setup-func) mode))
-                      (hook (or (get features 'setup-hook)
+                      (mode (or (get feature 'setup-mode) mode))
+                      (func (or (get feature 'setup-func) mode))
+                      (hook (or (get feature 'setup-hook)
                                 (get mode 'setup-hook)
                                 (intern (format "%s-hook" mode))))
-                      (map (or (get features 'setup-map)
+                      (map (or (get feature 'setup-map)
                                (get mode 'setup-map)
                                (intern (format "%s-map" mode))))))
                 body)
@@ -458,14 +555,32 @@ If FUNCTION is a list, apply BODY to all elements of FUNCTION."
 
 (setup-define :package
   (lambda (package)
-    `(unless (package-installed-p ',package)
-       (unless (memq ',package package-archive-contents)
-         (package-refresh-contents))
-       (package-install ',package)))
+    (let ((name (if (consp package)
+                    (progn
+                      (unless (fboundp 'package-vc-install)
+                        (error "Package-vc is not available"))
+                      (car package))
+                  package)))
+      `(unless (package-installed-p ',name)
+         (unless (assq ',package package-archive-contents)
+           (package-refresh-contents))
+         ,(cond
+           ((and (consp package) (consp (cdr package)))
+            `(package-vc-install ',package))
+           ((consp package) `(package-vc-install ',(car package)))
+           (t `(package-install ',package))))))
   :documentation "Install PACKAGE if it hasn't been installed yet.
-The first PACKAGE can be used to deduce the feature context."
+The first PACKAGE can be used to deduce the feature context.  If PACKAGE
+is a cons-cell (NAME . SPEC), then the NAME is installed using
+`package-vc-install' using the specification SPEC.  If SPEC is nil, then
+default to the package specification provided by ELPA."
   :repeatable t
-  :shorthand #'cadr)
+  :shorthand (lambda (package) (or (car-safe (cadr package)) (cadr package))))
+
+(setup-define :autoload-this
+  (lambda (&rest args)
+    `(autoload ',(setup-get 'func) ,(format "%s" (setup-get 'feature)) ,@args))
+  :documentation "Explicitly mark the context function to be autoloaded.")
 
 (setup-define :require
   (lambda (feature)
@@ -515,13 +630,18 @@ The first FEATURE can be used to deduce the feature context."
   :repeatable t)
 
 (setup-define :bind-into
-  (lambda (feature-or-map &rest rest)
-    (if (string-match-p "-map\\'" (symbol-name feature-or-map))
-        `(:with-map ,feature-or-map (:bind ,@rest))
-      `(:with-feature ,feature-or-map (:bind ,@rest))))
-  :documentation "Bind into keys into the map of FEATURE-OR-MAP.
+  (lambda (feature &rest rest)
+    (if (string-match-p "-map\\'" (symbol-name feature))
+        (progn
+          ;; https://lists.sr.ht/~pkal/public-inbox/%3C87pluzt28q.fsf@ushin.org%3E
+          ;; https://lists.sr.ht/~pkal/public-inbox/%3C87edbdma8j.fsf@ushin.org%3E
+          (warn "The `:bind-into' with a map %S is considered unreliable, and will be deprecated." feature)
+          `(:with-map ,feature (:bind ,@rest)))
+      `(:with-feature ,feature (:bind ,@rest))))
+  :documentation "Bind into keys into the map of FEATURE.
 The arguments REST are handled as by `:bind'."
   :debug '(sexp &rest form sexp)
+  :ensure '(nil &rest kbd func)
   :indent 1)
 
 (setup-define :hook
@@ -541,6 +661,13 @@ The arguments REST are handled as by `:bind'."
   :documentation "Add current function to HOOK."
   :repeatable t)
 
+(setup-define :bind-to
+  (lambda (binding)
+    `(global-set-key ,binding #',(setup-get 'func) ))
+  :documentation "Bind current function to HOOK."
+  :ensure '(kbd)
+  :repeatable t)
+
 (setup-define :option
   (setup-make-setter
    (lambda (name)
@@ -568,6 +695,17 @@ supported:
 (remove VAR)    Assuming VAR designates a list, remove all instances
                 of VAL.
 
+(append* VAR)  Assuming VAR designates a list, add each element
+               of VAL to the end of VAR, keeping their order,
+               unless it is already a member of the list.
+
+(prepend* VAR) Assuming VAR designates a list, add each element
+               of VAL to the start of VAR, keeping their order,
+               unless it is already a member of the list.
+
+(remove* VAR)  Assuming VAR designates a list, remove all
+               instances of each element of VAL.
+
 Note that if the value of an option is modified partially by
 append, prepend, remove, one should ensure that the default value
 has been loaded. Also keep in mind that user options customized
@@ -595,7 +733,18 @@ supported:
                 list.
 
 (remove VAR)    Assuming VAR designates a list, remove all instances
-                of VAL."
+                of VAL.
+
+(append* VAR)   Assuming VAR designates a list, add each element
+                of VAL to the end of VAR, keeping their order,
+                unless it is already a member of the list.
+
+(prepend* VAR)  Assuming VAR designates a list, add each element
+                of VAL to the start of VAR, keeping their order,
+                unless it is already a member of the list.
+
+(remove* VAR)   Assuming VAR designates a list, remove all
+                instances of each element of VAL."
   :debug '(sexp form)
   :repeatable t)
 
@@ -607,7 +756,7 @@ supported:
   :documentation "Add FUNCTION to HOOK only in buffers of the current mode."
   :debug '(symbolp sexp)
   :ensure '(nil func)
-  :repeatable t)
+  :repeatable '(1 . 1))
 
 (setup-define :also-load
   (lambda (feature)
@@ -662,9 +811,13 @@ yourself."
 
 (setup-define :and
   (lambda (&rest conds)
-    `(if (and ,@(butlast conds))
-         ,@(last conds)
-       ,(setup-quit)))
+    (let ((tail (car (last conds))))
+      `(if (and ,@(butlast conds))
+           ,@(cond
+              ((symbolp tail) '(nil))
+              ((consp tail) (last conds))
+              ((error "Illegal tail")))
+         ,(setup-quit))))
   :documentation "Abort evaluation of CONDS are not all true.
 The expression of the last condition is used to deduce the
 feature context."
@@ -672,31 +825,14 @@ feature context."
   (lambda (head)
     (unless (cdr head)
       (error ":and requires at least one condition"))
-    (let ((shorthand (get (caar (last head)) 'setup-shorthand)))
-      (and shorthand (funcall shorthand (car (last head))))))
+    (let ((tail (car (last head))))
+      (cond
+       ((symbolp tail) tail)
+       ((consp tail)
+        (let ((shorthand (get (car tail) 'setup-shorthand)))
+          (and shorthand (funcall shorthand tail)))))))
   :debug '(setup))
 
-
-;;; Obsoleted code
-
-(defun setup-ensure-kbd (sexp)
-  "Attempt to return SEXP as a key binding expression."
-  (cond ((stringp sexp) (kbd sexp))
-        ((symbolp sexp) `(kbd ,sexp))
-        (sexp)))
-(make-obsolete 'setup-ensure-kbd "Use :ensure keyword instead" "1.2.0")
-
-(defun setup-ensure-function (sexp)
-  "Attempt to return SEXP as a quoted function name."
-  (cond ((eq (car-safe sexp) 'function)
-         sexp)
-        ((eq (car-safe sexp) 'quote)
-         `#',(cadr sexp))
-        ((symbolp sexp)
-         `#',sexp)
-        (sexp)))
-(make-obsolete 'setup-ensure-function "Use :ensure keyword instead" "1.2.0")
-
 (provide 'setup)
 
 ;;; setup.el ends here
